!****************************************************************************
Module ConjGrad
!****************************************************************************

  Use TypeDef
  Use Normalisation
  Use Minimisation
  Use Initialisation
  Use Hamiltonian

  Implicit None

Contains

!****************************************************************************
  Subroutine CGDiagonalise(HamSize,HMatrix,NEig,EigValues,EigStates,tol,period,precond,conv)
!****************************************************************************

    integer, intent(in) :: HamSize,NEig,period,precond,conv
    real(dbl), intent(in) :: tol
    real(dbl), dimension(:) :: EigValues
    complex(dbl), dimension(:,:) :: HMatrix,EigStates

    type(EigSystemT), pointer :: EigSystem
    type(MinVecT), pointer :: MinVec,OldVec
    complex(dbl), dimension(:,:), allocatable :: H
    
    integer :: iband,cntr,oldband
    real(dbl) :: tolerance

    allocate(H(HamSize,HamSize))

    ! The size of HMatrix might be bigger than HamSize, with only the 
    ! HamSize * HamSize part filled
    Call ReadHMatrix(HMatrix,H,HamSize)

    Call MinVecAllocate(MinVec,HamSize)
    Call MinVecAllocate(OldVec,HamSize)

    Call EigSystemAllocate(EigSystem,HamSize,NEig)

!    do iband = 1, NEig

    cntr = 1
    iband = 1
    tolerance = tol

5   if(iband.le.NEig) then

       oldband = iband

       Call MinVecInit(MinVec,HamSize)
       Call MinVecInit(OldVec,HamSize)

       Call EigStateInit(EigSystem,iband)

       Call EigStateOrth(EigSystem,iband)

       Call EigenValues(EigSystem,H,iband)

       Call Diagonalise(HamSize,H,NEig,EigValues,EigStates, & 
       & iband,tolerance,precond,conv,EigSystem,MinVec,OldVec)

       iband = iband + 1

       !   If the calculation for the band has converged then iband has
       ! increased by one; if so then reset counter to 1 and tolerance to
       ! tol; else increase counter, if the counter crosses 10(or some 
       ! user defined number), then reset counter to 1 and decrease 
       ! tolerance. 
       if(oldband.ne.iband) then
          cntr = 1
          tolerance = tol
       else
          cntr = cntr + 1
          if(cntr.gt.period) then
             cntr = 1
             tolerance = 10*tolerance
             open(unit=110,file='Report.cg',status='old',action="write", & 
                  & position="append")
             write(110,*) 'Failed to converge in',period,' steps: Decreasing tolerance to ', tolerance
             close(unit=110)             
          end if
       end if

    end if

    if(iband.le.NEig) goto 5

    deallocate(H)
    Call EigSystemDestroy(EigSystem)
    Call MinVecDestroy(MinVec)
    Call MinVecDestroy(OldVec)

  end Subroutine CGDiagonalise

!****************************************************************************
  Subroutine ReadHMatrix(HMatrix,H,HamSize)
!***************************************************************************

    integer, intent(in) :: HamSize
    complex(dbl), dimension(:,:) :: HMatrix,H

    integer :: r,s

    do r = 1, HamSize
       H(r,r) = HMatrix(r,r)
       do s = r+1, HamSize
          H(r,s) = HMatrix(r,s)
          H(s,r) = conjg(HMatrix(r,s))
       end do
    end do

  end Subroutine ReadHMatrix

!***************************************************************************
  Subroutine Diagonalise(HamSize,HMatrix,NEig,EigValues,EigStates, & 
       & iband,tol,precond,conv,EigSystem,MinVec,OldVec)
!****************************************************************************

    ! Define the types of the subroutine arguments
    integer, intent(in) :: HamSize,NEig,precond,conv
    integer :: iband
    real(dbl), intent(in) :: tol
    real(dbl), dimension(:) :: EigValues
    complex(dbl), dimension(:,:) :: EigStates,HMatrix

    type(EigSystemT), pointer :: EigSystem
    type(MinVecT), pointer :: MinVec,OldVec

    integer :: iter
    real(dbl) :: resedue,tolsq

    tolsq = 10.d0*tol**2


    !   Iterate the Eigen Vector until convergence before moving over to the
    ! next higher band.
    do iter = 1, EigSystem%HSize

       !   Evaluate the Steepest Descent diection as the negative gradient
       ! In this case it is StDesc = -(H-lambda)|Psi>        
       Call StDescentVec(EigSystem,MinVec,iband) ! In "Minimise.f90"

       !   The resedue is the Steepst Descent Vector, so its square is 
       ! calculated, and the iterations are carried out until userdefined
       ! convergence is reached, OR 'HamSize' iterations have been carried out.
       resedue = Dot_Product(MinVec%StDesc,MinVec%StDesc)

       if(resedue.gt.tolsq) then

          !   Orhtogonalise the steepst descent vector against all the LOWER
          ! Eigen Vectors. (In "NormaliseMod.f90")
          Call LowerBandOrthogonalise(EigSystem,MinVec%StDesc,iband)

          !   PreCondition the Steepest Descent Vector In "HamiltonMod.f90"
          Call PreCondition(EigSystem,MinVec,HMatrix,iband)     
          
          !   Orthogonalise the PreConditioned Vector against all bands, lower 
          ! than and including the current band.  In "NormaliseMod.f90"
          Call BandOrthogonalise(EigSystem,MinVec%StDesc,iband) 
          
          !   Evaluate the conjugate direction orthogonal to the current band 
          ! and normalised. (In "MinimiseMod.f90")
          Call ConjDirection(EigSystem,MinVec,OldVec,iband)

          !   Do the line minimisation by searching for the energy minimum, and
          ! update the band.
          Call ExactEnergyMinimise(EigSystem,MinVec,HMatrix,iband)

       else
          
!          print *, 'Iterations:',iband,iter          
          goto 10
       end if

    end do
    
10  EigValues(iband) = EigSystem%Values(iband)
    EigStates(:,iband) = EigSystem%States(:,iband)

    
    ! Print the eigenvalue and resedue
    open(unit=110,file='Report.cg',status='old',action="write", & 
         & position="append")
    write(110,11) iband,EigValues(iband),sqrt(resedue)
    close(unit=110)

11     format(1X,'Band:',I2,1X,'EigenValues:',E15.5,1X,'Norm Resed:',E15.5)
12     format(1X,I4,E15.5)
13     format()




    if(resedue.gt.tolsq) then
       iband = iband -1
    end if


  end Subroutine Diagonalise
!*****************************************************************************



end Module ConjGrad

!*****************************************************************************
